#!/usr/bin/perl -w

# custom format: (maybe SET FORMATSTR=... in windows ? Perhaps more escaping also needed in .bat scripts)
# export FORMATSTR='c=%CLOCK MAP=%MAP MAT=%MAT stupido:$((%CLT*%MAT)) CLT=%CLT TPS=%TPS VBAT=%VBAT RPM=%RPM zzz'
# note the $(( expression to evaluate )) ... It is very powerful. (Would be sechole if FORMATSTR is used from web query: possibly usable to spawn arbitrary programs).
# Multiple $(( ... ))  $(( ... )) are allowed, but as a consequence, logically: 
# Warning: )) inside the expression is NOT allowed. Separate with spaces: ) )
$formatstr = $ENV{FORMATSTR} ? $ENV{FORMATSTR} :
  'CLOCK=%CLOCK MAP=%MAP MAT=%MAT CLT=%CLT TPS=%TPS VBAT=%VBAT RPM=%RPM';
$formatstr =~ s/ /\t/g; # space to tabs

# on-the-fly slurps vemslog as being written by VemsTune
# perl watch_vemslog.pl dir
# perl watch_vemslog.pl "/cygdrive/c/Users/Dave/Documents/VEMS_Files"
# so arg is just a dir under which we find newest dir/*/* fileabspos
# To avoid space-argument issues especially in windows,
# best to cd VEMS_Files and perl watch_vemslog.pl .
# Will find newest ./*/*.vemslog file (hopefully).

# 26kbyte: /svn/firmware/configexamples/KovacsAkos/BMW/S54/misclogs/v3.3_u009319-2014.10.21-18.29.47.vemslog
# 48 kbyte /svn/firmware/configexamples/KovacsAkos/BMW/S54/working/v3.3_u009469-2014.12.09-08.14.52.vemslog
# starting clock = 59 ... nice, so hex 3b (characteristic value, not 00 or 01) maybe soon goes to 3c ?
# ----
# 46kbyte: /svn/firmware/utilities/v3GUI-CommModel/logs/v3.2_n000315-2008.12.04-18.52.41.vemslog
# 1.3MByte: /svn/firmware/utilities/v3GUI-CommModel/logs/V3-log-2008.10.21-18.21.24.vemslog
# Holy Batman... These .vemslog files are apparently not of TF frames separated by 7e but rather
# "SD-card blocks" of 512 bytes, each containing 6 or 7 datasets ?
# so to locate RT data in vemslog
# 1) skip the first 512 bytes
# 2) search for 0A xx 07 at every 256 byte index
# 3) optional: search for sequence 4E 00 00 01 after that each 75 bytes is RT set
# 4) or just skip above steps and search for 4E 00 00 01 , if offset until next 4E 00 00 01 is 94 (dec) you found RT data
# no CRC checking, no advanced SPIR parsing, just getting the data from vemslog (easy way) ;)
# Dave Brul: i verified 0x3B is right after the 4E 00 00 01 sequence each time beginning of log.

my $RT_marker = sprintf( "%c%c%c%c", 0x4E, 0, 0, 0x01 ); # search for 4E 00 00 01 . RT data follows...
# in the old format eg CompilationTime=Mon Sep  3 15:10:58 CEST 2007
# ... DataLength=56
# ... Version=1.1.27
# eg. 1.2 MByte /svn/firmware/utilities/v3GUI-CommModel/logs/V3-log-2008.10.21-18.21.24.vemslog
# In that old format a runtime dataset starts every % 0x200 == 0 (0x200, 0x400, 0x600 ...). 
# Don't be tempted to search for 4e 00 00 0f that might not be useful.

my $dir='.' ; # 'C:\Users\Marcell\Documents\VEMS_Files'; # 'MarcellTest9-99876';
if( $ARGV[0] ){
	$dir=$ARGV[0];
}
# my $tmp="$basedir/$key";

my $str_sofar = '';
my $fileabspos = 0;	# absolute position in file (currently not tracked properly)
my $length_differs_from94_count = 0;
my $RPMstate=0; # 0: initial, nonzero otherwise. Could be: 1: 0 RPM, but was higher earlier 2: low RPM 10..200 3: normal/high RPM 255: crazy high RPM
my $RPMpeak=0;	# RPM peak so far
my $RPM=0;	# (prev) RPM value
# -------------------------
# finds and replaces $(( ... )), possible multiple occurances
sub RT_eval
{
  my $tstr = shift;
  my $replace = 'EVALTMPPLACEHOLDERR'; # any characteristic string would do (variable name style, without spaces)
  my $lreplace = length( $replace );
  my $expr4eval;
  my $max_eval=21; # don't allow more than this (or infloop if code is wrong)
  while($max_eval--){
      ($expr4eval) = ($tstr =~ /\$\(\((.+?)\)\)/ ); # NON-GREEDY match allows $(( %MAP*23 -1.2 )) even if multiple $(( )) expressions appear
      if(! $expr4eval){ last }; # more found => exit loop (note: last can exit from while loop, but NOT from do loop)
	my $evalresult = eval($expr4eval);
	# 62*23 plays bad in s/// so I better use index() and substr() C-style replacement, faster anyway
	my $lookfor = "\$(($expr4eval))";
	my $len1 = length( $lookfor );
	my $pos = index($tstr, $lookfor);
	while ( $pos > -1 ) {
	  substr( $tstr, $pos, $len1, $replace );
	  $pos = index( $tstr, $lookfor, $pos + $lreplace);
	}
	# avoid this: dangerous with "*" inside $expr4eval
	# $tstr =~ s/\$\(\($expr4eval\)\)/EVALTMPPLACEHOLDERR/ ; # temporary replacement (easier syntax and debugging ; some runtime cost)
	# $tstr =~ s/\$\(\($expr4eval\)\)/$evalresult/g;	# only first occurance or all ? (will be found again if really applicable)
	$tstr =~ s/$replace/$evalresult/g ;
	# print "WOW, eval:$expr4eval => $evalresult  tstr:$tstr END\n";
  }
  return $tstr;
}

# ---------
# uses RTFORMATSTR environment variable.
# RPM=%RPM ; TPS=%TPS ...
# would print RPM=4500 ; TPS = 83 ...
# s/// operator should do it ?
sub RT_print
{
	if( length( $formatstr ) < 3 ){
		return;
	}
  my $h = shift;
  # print " clock=$h->{CLOCK} ; TPS=$h->{TPS} ; RPM=$h->{RPM} ;\n";
  my $tstr = $formatstr;
  my @keys = qw(CLOCK STATUS MAP MAT CLT TPS VBAT RPM);
  my $k;
  foreach $k (@keys){
    my $v = $h->{$k};
    $tstr =~ s/\%$k/$v/g; # allow multiple replacement %TPS => TPS actual value
  }
  $tstr = RT_eval( $tstr );
  print $tstr, "\n";
}

# ----------------
# checks if RPM is roughly monotonous (increasing except at the end of triggerplay)
# updates $RPM
# Useful for bench testing, but DON'T call this when analyzing real engine data (comment RPM_validate() below).
sub RPM_validate
{
    my $h = shift;
    my $r = $h->{RPM};	# actual RPM; 

    my $rpm_error = 0;
    if( $r > $RPMpeak + 600 ){	# argh, RPM clibs too aggressively (or maybe to 25500)
      $rpm_error = 1;
    }
    if( $r < $RPMpeak - 420){ 	# $r dropped significantly
      # above 4800 it's OK... Once RPM will change from 5000 => 0 which is OK
      if( $RPMpeak > 1020 && $RPMpeak < 4820 ){	# this should not happen, maybe 900 => 480 is OK ?
	  $rpm_error = 1;
      }
    }
    if( $rpm_error && $RPMstate ){ # don't confuse
	  print "ERROR: RPM nonmonotonous (peak=$RPMpeak, $RPM=>$r), check TRIGGER... ==ERROR==\n";
    }
    if( $r > $RPMpeak || $r < 222){ # when $r falling back $RPMpeak should follow. We already fired any problems above.
      $RPMpeak = $r;
    }
    $RPMstate = 1;	# leave 0 state... could reflect RPM...
    $RPM = $r;
}
# ------
# maybe some config Hash should be used ?
# Perhaps a .pm should be shared with wx-serial/linux/testcase_process$ less process_after_anal1.pl
sub RT8
{
  return ord(substr( $_[0], $_[1], 1));
}
# 16 bit value (big endian)
sub RT16
{
  my ($RT_dataset, $idx) = @_;
  return RT8( $RT_dataset, $idx )*256 + RT8( $RT_dataset, $idx+1 ); # BIG endian
}
# -------------------
# extract RPM, wheelspeed and TPS
sub RT_process
{
    my $RT_frame = shift;
    my $hr = {};
    $hr->{CLOCK} = RT8( $RT_frame, 0 ); # TODO: clock could be interpolated for fraction of a second (determine rate of change, lookahead ?)
    # print "CLOCKALT=", ord(substr( $RT_frame, 0, 1));
    $hr->{STATUS} = RT8( $RT_frame, 2 );
    $hr->{MAP} = int( RT16( $RT_frame, 4 ) * 2.5 + 0.5) * 0.1 ; # kPa
    $hr->{MAT} = RT8( $RT_frame, 6 ) - 100;
    $hr->{CLT} = RT8( $RT_frame, 7 ) - 100;
    $hr->{TPS} = int( RT8( $RT_frame, 8 ) * 3.9215686 +0.499)*0.1; # 100 / 255 with some rounding
    $hr->{VBAT} = int( RT8( $RT_frame, 9 ) *300/255 + 0.5) * 0.1 ; # maybe * batt_cal / 255 also ? (config.batt_cal is around 192)
    $hr->{RPM} = RT16( $RT_frame, 14 ); # RPM is [14] *256 + [15]
    RT_print( $hr );
    RPM_validate( $hr );
}
# ---------
# verify length (94 is best, or slightly more if skipped a 0x200 boundary) 
sub RT_check
{
    my $RT_frame = shift;
    my $len = length( $RT_frame );
    if( $len < 150 ){
      if( 0 ){ # $fileabspos / 0x200 > $fileabspos_prev / 0x200 than boundary skipped, $len > 94 acceptable...
      
      }
    if( $len > 90 && $len < 100 ){	# 91 .. 99 should be == 94 (but sometimes 99 seem perfectly valid)
      if( $len != 94 ){
	  # print "WARNING! len=$len (not 94)\n";
	  $length_differs_from94_count++;
      }
      RT_process( $RT_frame );
    }
    }
}
# =======================
# TODO: modbus-CRC See crc_update_vector() in regressiontest_crctable.pl
# http://www.modbus.pl/node/23
# also see : http://www.modbustools.com/modbus_crc16.htm  ( uses an 256 entry (512 byte) flash-table to speed up CRC16 calculation )
sub modbus_check
{
  my $TF_frame = shift;
  # TODO: find and escape chr(0x7d) before modbus CRC !
  # for now only print length of $str_sofar
  print "\nTF_len=", length($TF_frame) ;
}

# find 7E ("HDLC-byte stuffing" triggerframe frames) and feeds escaped data into modbus CRC calculation
# This turned out to be USELESS in VT vemslog or SD log... Only useful for serial sniffed data.
sub TF_feed
{
  my $str = shift;
  my $pos;
  my $prevpos = 0;
  # process all TF frames:
  while( 0 < ($pos = index($str, chr(0x7e), $prevpos) )) { # -1 if not found
    $str_sofar .= substr( $str, $prevpos, $pos-$prevpos ); # cool, the last part of a TF frame
    modbus_check( $str_sofar );
    $str_sofar = '';
    $prevpos = $pos+1; # +1 ??? The 0x7E is not part of the next frame..
  }

  # append any leftover, remaining data:
  $str_sofar .= substr( $str, $prevpos );
}

# find most frequent offset of same byte's occurance:
# NOTE: this only works with full file, not fly-by-wire as VT writes the file
# found that after 1+93=94 bytes offset things tend to repeat
sub poor_mans_fft
{
  my $str = shift;
  # consider outer for loop:
  my $cval;
  foreach $cval ((0x5b) .. (0x7b)){ # reasonable for 1a .. fe (maybe not for 00..0a values)
    my $c = chr($cval);
    my $h = {};
    my $verystart = 0x200;
    my $prevpos = $verystart ; # anything before this is header: not interesting
    while( 0 < ($pos = index($str, $c , $prevpos) )) { # -1 if not found
      if( $prevpos > $verystart ){
	my $repeat_offset = $pos - $prevpos;
	$h->{ $repeat_offset } ++;
      }
      $prevpos = $pos+1; # +1 ??? The 0x7E is not part of the next frame..
    }
    # print the most frequent occurances (more than 10 ?)
    my $k;
    foreach $k (sort {$h->{$b} <=> $h->{$a}} (keys %$h)) {
      my $occu = $h->{$k};
      if( $occu > 5){
	print "$occu occurences of offset ($k) for chr($cval) decimal\n"; # bailout after a few entries...
      }
    }
  } # OUTER foreach
}
# ==============
# TODO: this code is "bogus" because if unluckily reading is always split in the middle of $RT_marker it won't be found
# Even if it does not happen, it should work if chars are fed 1 by 1 
sub RT_feed
{
  my $str = shift;
  my $pos;
  my $prevpos = 0;
  while( 0 < ($pos = index($str, $RT_marker, $prevpos) )) { # -1 if not found
    $str_sofar .= substr( $str, $prevpos, $pos+4-$prevpos ); # cool, the last part of RT frame
    RT_check( $str_sofar );
    $fileabspos += length( $str_sofar ); # TODO: this probably doesn't correctly track it...
    $str_sofar = '';
    $prevpos = $pos+4; # next frame cannot start earlier than this
  }

  # append any leftover, remaining data:
  $str_sofar .= substr( $str, $prevpos );
}
# ==============
# find newest file: updates $freshest->{name} and $freshest->{sec) if this file is modified more recently
# TODO: should we filter out vemsdetails.db ?
sub filecandidate
{
    my $file=shift;
    # print "FILECANDIDATE=$file\n" ;
    # ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size, $atime,$mtime,$ctime,$blksize,$blocks) = stat($filepath);
	$mtime_sec = (stat($file))[9]; # http://stackoverflow.com/questions/509576/how-do-i-get-a-files-last-modified-time-in-perl
	if( ! defined $freshest->{sec} || $mtime_sec > $freshest->{sec} ){
		$freshest->{sec} = $mtime_sec;
		$freshest->{name} = $file;
	 }
}
# ===============
sub slurp_file
{
	my $file = $freshest->{name};
    print "NEWEST=$file\n" ; 
    open FILE, '<' . $file || die $! ;
    binmode FILE ;
	while( $len = read(FILE, $data, 169108570)) {
		seek(FILE, 0, 1);	# clear EOF flag
		sleep(1);
		# print "read_len=$len\n" ;
		# TF_feed( $data );
		# poor_mans_fft( $data ); # this only works with full file, not fly-by-wire as VT writes the file
		RT_feed( $data );
		# push @array, $data ;
	} # while
	close FILE ;
}

# =============
opendir (TEMP1, "$dir") ;
my @dirs = readdir (TEMP1) ;
closedir (TEMP1) ;

foreach my $dirs (@dirs) { # *
    next if ($dirs eq "." || $dirs eq "..") ;
    # print "---->$dirs\n" ;
    my $d = "$dir/$dirs" ;
    if (-d "$d") { # */
        opendir (TEMP2, $d) || die $! ;
        my @files = readdir (TEMP2) ;
        closedir (TEMP2) ;

        foreach my $file (@files) { # */*
		my $f = "$d/$file" ;
		next if ( -d "$f" ); # filter out directories $file eq "." || $file eq ".." ||
            	# normal file
		filecandidate( $f ); # find newest file (eg. written in last 5 mins)
        }
     }
}

slurp_file();
#==========================
